//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "real_player.h"
#include <stdlib.h>
#include <algorithm>
#include "playerlist.h"
#include "armysetlist.h"
#include "FightDialog.h"
#include "stacklist.h"
#include "citylist.h"
#include <fstream>
#include "path.h"
#include "GameMap.h"
#include "army.h"
#include "hero.h"
#include "MoveResult.h"
#include <assert.h>
#include "signals.h"

using namespace std;

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl<<flush;}
#define debug(x)

RealPlayer::RealPlayer(string name, string armyset, SDL_Color color)
    :Player(name, armyset, color, Player::HUMAN)
{
}

RealPlayer::RealPlayer(string name, string armyset, SDL_Color color,
            Player::TYPE type)
    :Player(name, armyset, color, type)
{
}

RealPlayer::RealPlayer(const Player& player)
    :Player(player)
{
    d_type = HUMAN;
}

RealPlayer::RealPlayer(XML_Helper* helper)
    :Player(helper)
{
}

RealPlayer::~RealPlayer()
{
}

bool RealPlayer::startTurn()
{
    clearActionlist();

    return true;
}

bool RealPlayer::invadeCity(City* c)
{
    //for the realplayer, this function doesn't do a lot.
    //However, an AI player has to decide here what to do (occupy, raze,
    //pillage)
    sinvadingCity.emit(c);
    soccupyingCity.emit(c);
    return true;
}

bool RealPlayer::recruitHero(Hero* hero, int cost)
{
    // for the realplayer, this function also just raises a signal and looks
    // what to do next. The derived classes or W_Edit actually react on the
    // signal.

    return srecruitingHero.emit(hero, cost);
}

bool RealPlayer::stackSplit(Stack* s)
{
    debug("RealPlayer::stackSplit("<<s->getId()<<")")

    Army* ungrouped = s->getFirstUngroupedArmy();
    if (!ungrouped)        //no armies to split
    {
        return false;
    }

    bool all_ungrouped = true;    //the whole army would be split
    for (Stack::iterator it = s->begin(); it != s->end(); it++)
    {
        if ((*it)->isGrouped())
        {
            all_ungrouped = false;
        }
    }
    if (all_ungrouped)
    {
        return false;
    }

    Stack* new_stack = new Stack(this, s->getPos());

    while (ungrouped)
    {
        new_stack->push_back(ungrouped);
        s->erase(find(s->begin(), s->end(), ungrouped));

        ungrouped->setGrouped(true);
        ungrouped = s->getFirstUngroupedArmy();
    }

    d_stacklist->push_back(new_stack);

    Action_Split* item = new Action_Split();
    item->fill_data(s, new_stack);
    d_actions.push_back(item);

    return true;
}

bool RealPlayer::stackJoin(Stack* receiver, Stack* joining)
{
    debug("RealPlayer::stackJoin("<<receiver->getId()<<","<<joining->getId()<<")")

    if ((receiver == 0) || (joining == 0))
        return false;

    if ((receiver->size() + joining->size()) > 8)
        return false;

    for (Stack::iterator it = joining->begin(); it != joining->end(); it++)
    {
        receiver->push_back(*it);
        (*it)->setGrouped(true);
    }

    Action_Join* item = new Action_Join();
    item->fill_data(receiver, joining);
    d_actions.push_back(item);

    joining->clear();    //clear only erases the pointers not the armies
    d_stacklist->flRemove(joining);
    supdatingStack.emit(0);

    return true;
}


bool RealPlayer::stackMove(Stack* s)
{
    debug("player::stackMove(Stack*)")

    if (s->getPath()->empty())
    {
        return false;
    }

    Path::iterator it = s->getPath()->end();
    it--;
    MoveResult *result = stackMove(s, **(it), true);
    bool ret = result->moveSucceeded();
    delete result;
    result = 0;
    return ret;
}

MoveResult *RealPlayer::stackMove(Stack* s, PG_Point dest, bool follow)
{
    debug("player::stack_move()");
    //if follow is set to true, follow an already calculated way, else
    //calculate it here
    if (!follow)
    {
        s->getPath()->calculate(s, dest);
    }
    else
    {
        s->getPath()->checkPath(s);
    }

    if (s->getPath()->empty())
    {
        return new MoveResult(false);        //way to destination blocked
    }

    int stepCount = 0;
    while ((enoughMoves(s)) && (s->getPath()->size() > 1))
    {
        if (stackMoveOneStep(s)) stepCount++;
        supdatingStack.emit(0);
    }

    if ((s->getPath()->size() == 1) && enoughMoves(s))
    //now look for fight targets, joins etc.
    {
        PG_Point pos = **(s->getPath()->begin());
        City* city = Citylist::getInstance()->getObjectAt(pos);
        Stack* target = Stacklist::getObjectAt(pos);

        //first fight_city to avoid ambiguity with fight_army
        if (city && (city->getPlayer() != this) && (!city->isBurnt()))
        {
            Fight::RESULT result;
            MoveResult *moveResult = new MoveResult(true);
            vector<Stack*> def_in_city = Stacklist::defendersInCity(city);
            if (!def_in_city.empty())
            {
                // This is a hack to circumvent the limitations of stackFight.
                // Create a dummy stack at the target position if neccessary
                // and start a fight with this dummy stack.
                if (!target)
                    target = new Stack(city->getPlayer(), pos);
                result = stackFight(&s, &target, false);
                if (target && target->empty())
                    delete target;
            }
            else
                result = Fight::ATTACKER_WON;

            moveResult->setFightResult(result);

            // We may only take the city if we have defeated all defenders
            if (Stacklist::defendersInCity(city).empty())
            {
                if (stackMoveOneStep(s))
                    stepCount++;
                invadeCity(city); //determine what to do with city
            }
            else if (s)
            {
                // we didn't suceed in defeating the defenders, but at least our
                // stack has survived. Subtract some MP because fighting takes
                // time.
                s->decrementMoves(2);
            }
            
            moveResult->setStepCount(stepCount);
            supdatingStack.emit(0);
            return moveResult;
        }
        
        //another friendly stack => join it
        else if (target && target->isFriend(this))
        {
            if (stackMoveOneStep(s)) stepCount++;
            stackJoin(Stacklist::getAmbiguity(s), s);
            supdatingStack.emit(0);
            MoveResult *moveResult = new MoveResult(true);
            moveResult->setStepCount(stepCount);
            moveResult->setJoin(true);
            return moveResult;
        }
        
        //enemy stack => fight
        else if (target)
        {
            MoveResult *moveResult = new MoveResult(true);
        
            Fight::RESULT result = stackFight(&s, &target, false);
            moveResult->setFightResult(result);
            if (!target)
            {
                if (stackMoveOneStep(s))
                    stepCount++;
            }
            else if (s)
                s->decrementMoves(2);
            
            supdatingStack.emit(0);
            moveResult->setStepCount(stepCount);
            return moveResult;
        }
        
        //else
        if (stackMoveOneStep(s)) stepCount++;
        supdatingStack.emit(0);
        MoveResult *moveResult = new MoveResult(true);
        moveResult->setStepCount(stepCount);
        return moveResult;
    }

    //If there is another stack where we landed, join it. We can't have two
    //stacks share the same maptile
    if (Stacklist::getAmbiguity(s))
    {
        stackJoin(Stacklist::getAmbiguity(s), s);
        supdatingStack.emit(0);
        MoveResult *moveResult = new MoveResult(true);
        moveResult->setStepCount(stepCount);
        moveResult->setJoin(true);
        return moveResult;
    }

    return new MoveResult(true);
}


Fight::RESULT RealPlayer::stackFight(Stack** attacker, Stack** defender, bool ruin) 
{
    debug("stackFight: player = " << getName()<<" at position "
          <<(*defender)->getPos().x<<","<<(*defender)->getPos().y);

    // save the defender's player for future use
    Player* pd = (*defender)->getPlayer();

    // I suppose, this should be always true, but one can never be sure
    bool attacker_active = ((*attacker) == d_stacklist->getActivestack());
    
    // Fighting goes in three steps. First, we collect all units around the
    // defender which belong either to the attacking or the defending player.
    // Then we actually do the fight, and finally, we remove the dead,
    // distribute the xp, raise levels etc.

    // First, collect the armies. Don't do this if we search a ruin!!

    vector<Stack*> attackers;
    vector<Stack*> defenders;

    if (ruin)
    {
        attackers.push_back(*attacker);
        defenders.push_back(*defender);
    }
    else
    {
        PG_Point rel_pos;
        PG_Point pos = **((*attacker)->getPath()->begin());
        //use all stacks adjacent to defender.
        for (int i = -1; i < 2; i++)
        {
            for (int j = -1; j < 2; j++)
            {
                rel_pos.x = pos.x+i;
                rel_pos.y = pos.y+j;
                Stack* army = Stacklist::getObjectAt(rel_pos);
                
                //is friendly to defender.
                if (army && army->isFriend((*defender)->getPlayer())) 
                {
                    defenders.push_back(army);
                }
                
                //is friendly to attacker.
                if (army && army->isFriend((*attacker)->getPlayer()))
                {
                    attackers.push_back(army);
                }
                
                // we don't care about a 3rd player?
                // UL: No, the 3rd player's armies have no reason to join the fight.
            }
        }
    }
    
    
    // Second, do the fighting
    
    if (attackers.empty() || defenders.empty())
        return Fight::ATTACKER_WON;

    //the timers of bigmap and smallmap cause flickering of the screen, so
    //we remove them for the duration of the fight
    sinterruptTimers.emit();

    // Initialize and do the Fight
    FightDialog *fdlg= new FightDialog(attackers, defenders, 0, ruin);
    fdlg->Show();
    fdlg->battle();

    Fight::RESULT result = fdlg->getResult();
    delete fdlg;
    //add result to actionlist


    // And do the cleanup
    
    culpritsType attackerHeroes, defenderHeroes;

    // get attacker and defender heroes
    getHeroes(attackers, attackerHeroes);
    getHeroes(defenders, defenderHeroes);

    // here we calculate also the total XP to add when a player have a battle
    // clear dead defenders
    debug("clean dead defenders");
    double defender_xp = removeDeadArmies(defenders, attackerHeroes);

    // and dead attackers
    debug("clean dead attackers");
    double attacker_xp = removeDeadArmies(attackers, defenderHeroes);

    debug("restart timers");
    //don't forget to restart the timers
    scontinueTimers.emit();

    // after a fight: if the attacker's stack won - emit
    // supdatingStack signal, else emit sdyingStack

    debug("after fight: attackers empty? " << attackers.empty()
          << "(" << attackers.size() << ")");


    //only emit this signal when the defender has not won!
    if (!attackers.empty())
    {
        debug("attacker won: supdatingStack");
        supdatingStack.emit(0);
        if (defender_xp != 0)
            updateArmyValues(attackers, true, defender_xp);
    }
    else
    {
        debug("attacker lost: sdyingStack");
        sdyingStack.emit(0);
    }
    if (attacker_xp != 0)
        updateArmyValues(defenders, false, attacker_xp);

    // Set the attacker and defender stack to 0 if neccessary. This is a great
    // help for the functions calling stackFight (e.g. if a stack attacks
    // another stack and destroys it without winning the battle, it may take the
    // position of this stack)

    // First, the attacker...
    bool exists = false;
    for (Stacklist::iterator it = d_stacklist->begin(); it != d_stacklist->end(); it++)
        if ((*it) == (*attacker))
            exists = true;
    if (!exists)
    {
        (*attacker) = 0;
        if (attacker_active)
            d_stacklist->setActivestack(0);
    }

    // ...then the defender.
    exists = false;
    if (pd)
        for (Stacklist::iterator it = pd->getStacklist()->begin();
             it != pd->getStacklist()->end(); it++)
        {
            if ((*it) == (*defender))
                exists = true;
        }
    else
        exists = true;
    if (!exists)
        (*defender) = 0;

    
    return result;
}


void RealPlayer::updateArmyValues(vector<Stack*>& stacks,bool emit,double xp_sum)
{
    vector<Stack*>::iterator it;

    for (it = stacks.begin(); it != stacks.end(); )
    {
        debug("Stack: " << (*it))
        double numberarmy= (*it)->size();

        for (Stack::iterator sit = (*it)->begin(); sit != (*it)->end();)
        {
            debug("Army: " << (*sit))

            // here we adds XP
            (*sit)->gainXp((double)((xp_sum)/numberarmy));
            debug("Army gets " << (double)((xp_sum)/numberarmy) << " XP")

            // here we adds 1 to number of battles
            (*sit)->setBattlesNumber((*sit)->getBattlesNumber()+1);
            debug("Army battles " <<  (*sit)->getBattlesNumber())

            if(((*sit)->getBattlesNumber())>10 && !((*sit)->getMedalBonus(2)))
            {
                (*sit)->setMedalBonus(2,true);
                // We must recalculate the XPValue of this unit since it got a medal
                (*sit)->setXpReward((*sit)->getXpReward()+1);
                // We get the medal bonus here
                (*sit)->setOrigMaxHp((*sit)->getOrigMaxHp()+4);
                // Emit only if the player is Human
                if (emit) snewmedalArmy.emit(*sit);
            }

            debug("Army percentage of hits " <<  (double)(*sit)->getNumberHasHit()/(double)(*sit)->getNumberAttacks()*(double)100)
            debug("number of hits " <<  (*sit)->getNumberHasHit())
            debug("number of attacks " <<  (*sit)->getNumberAttacks())

            // Only give medals if the unit has attacked often enough, else
            // medals loose the flair of something special
            if(((double)(*sit)->getNumberHasHit()/(double)(*sit)->getNumberAttacks())>0.9
                 && !(*sit)->getMedalBonus(0) && (*sit)->getNumberAttacks() >= 10)
            {
                (*sit)->setMedalBonus(0,true);
                // We must recalculate the XPValue of this unit since it got a medal
                (*sit)->setXpReward((*sit)->getXpReward()+1);
                // We get the medal bonus here
                (*sit)->setOrigStrength((*sit)->getOrigStrength()+1);
                // Emit only if the player is Human
                if (emit) snewmedalArmy.emit(*sit);
            }

            debug("Army percentage of been hit " <<  (double)(*sit)->getNumberHasBeenHit()/(double)(*sit)->getNumberAttacks()*(double)100)
            debug("number of been hit " <<  (*sit)->getNumberHasBeenHit())

            // see above about a minimum level of the fight
            if(((double)(*sit)->getNumberHasBeenHit()/(double)(*sit)->getNumberAttacks())<0.1
                && !(*sit)->getMedalBonus(1) && (*sit)->getNumberAttacks() >= 10)
            {
                (*sit)->setMedalBonus(1,true);
                // We must recalculate the XPValue of this unit since it got a medal
                (*sit)->setXpReward((*sit)->getXpReward()+1);
                // We get the medal bonus here
                (*sit)->setOrigDefense((*sit)->getOrigDefense()+1);
                // Emit only if the player is Human
                if (emit) snewmedalArmy.emit(*sit);
            }

            for(int i=0;i<3;i++)
            {
                debug("MEDAL[" << i << "]==" << (*sit)->getMedalBonus(i))
            }

            // We reset the hit values after the battle
            (*sit)->setNumberHasHit(0);
            (*sit)->setNumberHasBeenHit(0);
            (*sit)->setNumberAttacks(0);

            while((*sit)->canGainLevel())
            {
                // Units not associated to a player never raise levels.
                if ((*sit)->getPlayer() == 0)
                    break;

                //Here this for is to check if army must raise 2 or more levels per time
                //depending on the XP and level itsself

                debug("ADVANCING LEVEL "<< "CANGAINLEVEL== " << (*sit)->canGainLevel())
                if (emit)
                {
                    //This is the human player army that advance level!
                    debug("Army Gained Level!")
                    debug("XP--> " << (*sit)->getXP()  << "  LEVEL== " << (*sit)->getLevel())
                    debug("Signal gained  Level!")
                    snewlevelArmy.emit(*sit);
                }
                else
                {
                    //This is the computer army that advance level!
                    debug("AI Army Gained Level!")
                    debug("XP--> " << (*sit)->getXP()  << "  LEVEL== " << (*sit)->getLevel())
                    debug("Signal gained  Level!")
                    snewlevelAiArmy.emit(*sit);
                }

                // We must recalculate the XPValue of this unit since it advanced a level.
                (*sit)->setXpReward((*sit)->getXpReward()+1);

                // here the army gains a level
                (*sit)->gainLevel();
            }
            debug("Army new XP=" << (*sit)->getXP())
            sit++;
        }
        it++;
    }
}

void RealPlayer::getHeroes(vector<Stack*>& stacks, culpritsType& dst)
{
    vector<Stack*>::iterator it;
    for (it = stacks.begin(); it != stacks.end(); it++)
        (*it)->getHeroes(dst);
}

double RealPlayer::removeDeadArmies(vector<Stack*>& stacks,
                                  culpritsType& culprits)
{
    double total=0;
    Player *owner = NULL;
    if (stacks.empty() == 0)
    {
        owner = stacks[0]->getPlayer();
        debug("Owner = " << owner);
        if (owner)
            debug("Owner of the stacks: " << owner->getName()
                  << ", his stacklist = " << owner->getStacklist());
    }
    for (unsigned int i = 0; i < culprits.size(); i++)
        debug("Culprit: " << culprits[i]);

    vector<Stack*>::iterator it;
    for (it = stacks.begin(); it != stacks.end(); )
    {
        debug("Stack: " << (*it))
        for (Stack::iterator sit = (*it)->begin(); sit != (*it)->end();)
        {
            debug("Army: " << (*sit))
            if ((*sit)->getHP() <= 0)
            {
                debug("Army: " << (*sit)->getName())
                debug("Army: " << (*sit)->getXpReward())
                //Add the XP bonus to the total of the battle;
                total+=(*sit)->getXpReward();
                // here we destroy the army, so we send
                // the signal containing the fight data
                debug("sending sdyingArmy!")
                sdyingArmy.emit(*sit, culprits);
                sit = (*it)->flErase(sit);
                continue;
            }
            sit++;
        }
        
        debug("Is stack empty?")
            
        if ((*it)->empty())
        {
            if (owner)
            {
                debug("Removing this stack from the owner's stacklist");
                owner->deleteStack(*it);
            }
            else // there is no owner - like for the ruin's occupants
                debug("No owner for this stack - do stacklist too");

            debug("Removing from the vector too (the vetor had "
                  << stacks.size() << " elt)");
            it = stacks.erase(it);
        }
        else
            it++;
    }
    debug("after removeDead: size = " << stacks.size());
    return total;
}

bool RealPlayer::stackSearchRuin(Stack* s, Ruin* r)
{
    debug("RealPlayer::stack_search_ruin")

    //throw out impossible actions
    if ((s->getPos().x != r->getPos().x) ||
        (s->getPos().y != r->getPos().y))
    {
        cerr <<_("Error: searching stack and ruin to be searched not on same position\n");
        exit(-1);
    }

    if (r->isSearched())
        return false;

    Stack* keeper = r->getOccupant();

    stackFight(&s, &keeper, true);

    // special situation - stackFight doesn't save history
    // when there were no defenders
    if (keeper == NULL)
    {
        Action_Search* item = new Action_Search();
        item->fill_data(r, 0, 0, 0, true);
        d_actions.push_back(item);
    }

    // did the explorer lose?
    if (keeper && keeper->size())
        return false;

    r->setOccupant(0);
    if (keeper)
        delete keeper;

    // but if the hero was killed, do not emit the signal about
    // search
    Army* hero = s->getFirstHero();
    if (hero)
    {
       debug("emiting signal - ruin searched ");
        // emit the signal - (ruin_id, hero_id)
       sruinSearched.emit(r->getId(), hero->getId());
    }

    //The fight has been done or left out, now comes the reward. Up to now,
    //we only give some gold (automatically done with give_reward())
    r->setSearched(true);
    ssearchingRuin.emit(r);

    // do we reward a player event when his hero died?
    giveReward(1, s);
    supdatingStack.emit(0);
    return true;
}

bool RealPlayer::stackVisitTemple(Stack* s, Temple* t)
{
    debug("RealPlayer::visit_temple")

    //abort in case of impossible action
    if ((t->getPos().x != s->getPos().x)
        || (t->getPos().y != s->getPos().y))
    {
        cerr <<_("Stack tried to visit temple at wrong location\n");
        exit(-1);
    }

    // you have your stack blessed (+1 strength)
    s->bless();

    Action_Temple* item = new Action_Temple();
    item->fill_data(t, s);
    d_actions.push_back(item);

    svisitingTemple.emit(t);
    
    supdatingStack.emit(0);

    return true;
}

bool RealPlayer::cityOccupy(City* c)
{
    debug("cityOccupy")

    c->conquer(this);

    //set the production to the cheapest armytype
    c->setProduction(-1, false);
    if (c->getArmytype(0, false) != -1)
        c->setProduction(0, false);

    Action_Occupy* item = new Action_Occupy();
    item->fill_data(c);
    d_actions.push_back(item);

    //to signal that the cities have changed a bit
    supdatingCity.emit(c);

    return true;
}

bool RealPlayer::cityPillage(City* c)
{
    debug("RealPlayer::cityPillage")

    Action_Pillage* item = new Action_Pillage();
    item->fill_data(c);
    d_actions.push_back(item);
    
    // pillaging works in two steps

    // if the city has been upgraded, reduce the level and get some of the gold
    if (c->reduceDefense())
    {
        addGold((c->getDefenseLevel()+1) * 200);
        cityOccupy(c);
        return true;
    }

    // if demolishing the city is not possible, remove all production (at this
    // level the city only has two, so this is not as hard as it seems)

    for (int i = 0; i < c->getNoOfBasicProd(); i++)
        if (c->getArmytype(i, false) != -1)
        {
            addGold(50);
            c->removeBasicProd(i);
        }

    cityOccupy(c);
    return true;
}

bool RealPlayer::cityRaze(City* c)
{
    debug("RealPlayer::cityRaze")

    Action_Raze* item = new Action_Raze();
    item->fill_data(c);
    d_actions.push_back(item);

    c->setBurnt(true);
    //make sure there are no side effects e.g. when counting the cities of
    //a player
    c->conquer(Playerlist::getInstance()->getNeutral());

    supdatingCity.emit(c);

    return true;
}

bool RealPlayer::cityUpgradeDefense(City* c)
{
    debug("RealPlayer::city_upgrade_defense")

    int needed_gold = c->getDefenseLevel() * 1000;

    if ((needed_gold >= 3000) || (d_gold < needed_gold))
        return false;    //maximum level is 3 or not enough gold

    c->raiseDefense();
    withdrawGold(needed_gold);

    Action_Upgrade_Defense* item = new Action_Upgrade_Defense();
    item->fill_data(c);
    d_actions.push_back(item);

    return true;
}

bool RealPlayer::cityBuyProduction(City* c, int slot, int type, bool advanced)
{
    Armyset* as;

    if (advanced)
        as = getArmyset();
    else
        as = Armysetlist::getInstance()->getArmyset("default");

    // sort out unusual values (-1 is allowed and means "scrap production")
    if ((type < -1) || (type >= as->getSize()))
        return false;
    
    // return if we don't have enough money
    if ((type != -1) && (as->army[type].getProductionCost() > d_gold))
        return false;

    // return if the city already has the production
    if (c->hasProduction(type, as))
        return false;

    // now we assume everything is ok. Remove old production and set the new one
    if (advanced)
    {
        c->removeAdvancedProd(slot);
        if (!c->addAdvancedProd(slot, type))
            return false;
    }
    else
    {
        c->removeBasicProd(slot);
        if (!c->addBasicProd(slot, type))
            return false;
    }
    
    // and do the rest of the neccessary actions
    withdrawGold(as->army[type].getProductionCost());

    Action_Buy_Production* item = new Action_Buy_Production();
    item->fill_data(c, type);
    d_actions.push_back(item);

    return true;
}

bool RealPlayer::cityChangeProduction(City* c, int slot, bool advanced)
{
    c->setProduction(slot, advanced);

    Action_Change_Production* item = new Action_Change_Production();
    item->fill_data(c, slot);
    d_actions.push_back(item);

    return true;
}

bool RealPlayer::giveReward(int level, Stack* s, fl_Treasure* reward)
{
    debug("RealPlayer::give_reward")

    //by now we don't have the infrastructure to give something different
    //from gold, so this function isn't great

    int gold_amount = rand() % 1000;
    addGold(gold_amount);

    fl_Treasure treasure;
    treasure.type = TREASURE_GOLD;
    treasure.subtype = 0;
    treasure.amount = gold_amount;

    Action_Reward* item = new Action_Reward();
    item->fill_data(0, treasure);
    d_actions.push_back(item);

    return true;
}

bool RealPlayer::stackMoveOneStep(Stack* s)
{
    if (!s)
        return false;
    
    if (!enoughMoves(s))
        return false;

    PG_Point dest = *(s->getPath()->front());

    int needed_moves = GameMap::getInstance()->getTile(dest.x,dest.y)->getMoves();
    unsigned int maptype = GameMap::getInstance()->getTile(dest.x,dest.y)->getMaptileType();
    City* c = Citylist::getInstance()->getObjectAt(dest.x, dest.y);

    for (Stack::iterator it = s->begin(); it != s->end(); it++)
    //calculate possible move boni for each army
    {
        if (c != 0)
        //cities cost one MP
        {
            (*it)->decrementMoves(1);
            continue;
        }

        if ((*it)->getMoveBonus() == maptype)
        //if army has move bonus, it takes only 2 move points...
        {
            (*it)->decrementMoves(2);
            continue;
        }
        //else the whole
        (*it)->decrementMoves(needed_moves);
    }

    s->moveOneStep();    //this is only for updating positions etc.
    d_fogmap->alterFogRadius(dest, s->getMaxSight(), FogMap::OPEN);

    //! signal that we have moved the stack
    smovingStack.emit(s);

    Action_Move_Step* item = new Action_Move_Step();
    item->fill_data(s, dest);
    d_actions.push_back(item);

    return true;
}

// End of file
